#ifndef __PARSE_H
#define __PARSE_H

#include <stxxl/bits/containers/sorter.h>
#include <stxxl/bits/containers/stack.h>

#include "radixsort.h"
#include "utils.h"
#include "uint40.h"
#include "stream.h"

#define BLOCK_SIZE (1024 * 1024)
#define UNDEFINED 549755813887L

struct PsvNsvTriple {
  PsvNsvTriple() {}
  PsvNsvTriple(uint40 idx, uint40 psv, uint40 nsv)
    : m_idx(idx), m_psv(psv), m_nsv(nsv) {
  }

  uint40 m_idx, m_psv, m_nsv;
};

struct CmpPsvNsvTriple {
  inline bool operator() (const PsvNsvTriple &a, const PsvNsvTriple &b) const {
    return a.m_idx < b.m_idx;
  }

  PsvNsvTriple min_value() const {
    return PsvNsvTriple(std::numeric_limits<uint40>::min(), 0, 0);
  }

  PsvNsvTriple max_value() const {
    return PsvNsvTriple(std::numeric_limits<uint40>::max(), 0, 0);
  }
};

namespace stxxl {
namespace algorithm {

template <typename Iterator>
struct Sort<Iterator, CmpPsvNsvTriple> {
  static inline void sort(Iterator begin, Iterator end, CmpPsvNsvTriple& cmp) {
    radixsort8msb_copy2(begin, end, sizeof(uint40)-1, cmp);
  }
};

} // namespace algorithm
} // namespace stxxl

inline long max_alloc() {
  return stxxl::block_manager::get_instance()->max_allocated();
}

// Compute phrase starting at position i, given nsv = NSV_text[i] and
// psv = PSV_text[i] and store the result in (pos, len).
//
// Returns: the starting position of the next phrase.
inline long parse_phrase(unsigned char *text, long n, long i,
    long psv, long nsv, long &pos, long &len) {
  pos = len = 0L;

  if (nsv == UNDEFINED) {
    while (text[psv + len] == text[i + len])
      ++len;

    pos = psv;
  } else if (psv == UNDEFINED) {
    while (i + len < n && text[nsv + len] == text[i + len])
      ++len;

    pos = nsv;
  } else {
    while (text[psv + len] == text[nsv + len])
     ++len;

    if (text[i + len] == text[psv + len]) {
      ++len;
      while (text[i + len] == text[psv + len])
        ++len;
      pos = psv;
    } else {
      while (i + len < n && text[i + len] == text[nsv + len])
        ++len;
      pos = nsv;
    }
  }
  
  if (len == 0L)
    pos = text[i];

  return i + std::max(1L, len);
}

// Parse given file using at most "ram_use" bytes.
//
// If "out_filename" is nonempty, the resulting parsing is written there.
// The parsing is encoded as a sequence of pairs (pos, len), where
//   - pos is the starting position of the source
//   - len is the phrase length
//  or the phrase is (c, 0), where c is the symbol from the text.
void parse_file(std::string in_fname, long ram_use, std::string out_fname) {
  long double start = utils::wclock();

  long length = utils::file_size(in_fname);
  long double length_MiB = (long double)length / (1 << 20);
  long double ram_use_MiB = (long double)ram_use / (1 << 20);

  fprintf(stderr, "Text filename = %s\n", in_fname.c_str());
  fprintf(stderr, "Text length = %ld\n", length);
  fprintf(stderr, "Memory budget = %.2LfMiB\n", ram_use_MiB);

  stream_reader<uint40> *sa = new stream_reader<uint40>(in_fname + ".sa5");

  CmpPsvNsvTriple cmp_psvnsv;
  long creator_budget = (24L * std::max(length, ram_use)) / 37L;

  long merger_budget;
  if (ram_use - length < (50L << 20)) {
    fprintf(stderr, "Error: insufficient memory! The algorithm requires "
        "at least RAM to hold text and about 50MiB for external memory sorting.\n");
    std::exit(EXIT_FAILURE);
  }

  // Do the merging while the input text is in memory.
  merger_budget = std::max((160L << 20), ram_use - length);

  fprintf(stderr, "Run creator budget = %.2LfMiB\n",
      (long double)creator_budget / (1 << 20));
  fprintf(stderr, "Run merger budget = %.2LfMiB\n",
      (long double)merger_budget / (1 << 20));

  stxxl::sorter<PsvNsvTriple, CmpPsvNsvTriple, BLOCK_SIZE>
      sorter(cmp_psvnsv, creator_budget, merger_budget);

  typedef stxxl::STACK_GENERATOR<long>::result stack_type;
  stack_type *stack = new stack_type();

  long double nsvpsv_start = utils::wclock();
  for (long i = 0, dbg = 0; i < length; ++i, ++dbg) {
    long SAi = sa->read().ll();

    while (!(stack->empty()) && stack->top() > SAi) {
      long cur = stack->top();
      stack->pop();

      long nsv = SAi;
      long psv = (stack->empty() ? UNDEFINED : (stack->top()));
      sorter.push(PsvNsvTriple(cur, psv, nsv));
    }

    stack->push(SAi);

    if (dbg == (1 << 23)) {

      dbg = 0;
      long double elapsed = utils::wclock() - nsvpsv_start;
      long double scanned_m = i / (1024.L * 1024);
      long inp_vol = sizeof(uint40) * i;
      long out_vol = sizeof(PsvNsvTriple) * sorter.size();
      long tot_vol = inp_vol + out_vol;
      long double tot_vol_m = tot_vol / (1024.L * 1024);
      long double io_speed = tot_vol_m / elapsed;
      fprintf(stderr, "Computing NSV/PSV: scanned %.1LfM SA values (%.1Lf%%), "
          "time = %.1Lfs (%.3Lfs/iMiB), io = %.2LfMiB/s\r",
          scanned_m, (100.L * i) / length, elapsed, elapsed / scanned_m,
          io_speed);
    }
  }

  while (!(stack->empty())) {
    long cur = stack->top();
    stack->pop();
    long nsv = UNDEFINED;
    long psv = ((stack->empty()) ? UNDEFINED : (stack->top()));
    sorter.push(PsvNsvTriple(cur, psv, nsv));
  }
  
  delete stack;
  delete sa;
  sorter.sort();
  
  long double nsvpsv_time = utils::wclock() - nsvpsv_start;
  fprintf(stderr, "\nComputed NSV/PSV in %.1Lfs (%.3Lfs/iMiB), "
      "max-alloc = %.2Lfbytes/char\n",
      nsvpsv_time, nsvpsv_time / length_MiB, 1.L * max_alloc() / length);

  fprintf(stderr, "Reading text: ");
  long double reading_start = utils::wclock();
  unsigned char *text;
  utils::read_file<unsigned char>(text, length, in_fname);
  long double reading_time = utils::wclock() - reading_start;
  fprintf(stderr, "%.2Lfs (%.4Lfs/iMiB)\n\n", reading_time,
      reading_time / length_MiB);


  fprintf(stderr, "Parsing:\r");
  long double parse_start = utils::wclock();

  bool write_out = (out_fname.empty() == false);
  stream_writer<uint40> *out_writer = NULL;
  if (write_out)
    out_writer = new stream_writer<uint40>(out_fname);

  long next = 1, pos, len, nphrase = 0, out_vol = 0;

  nphrase = 1;
  if (write_out) {
    out_writer->write((long)text[0]);
    out_writer->write(0L);
    out_vol += 2 * sizeof(uint40);
  }

  for (long i = 0, dbg = 0; i < length; ++i, ++dbg) {
    if (next == i) {
      ++nphrase;
      PsvNsvTriple t = *sorter;

      long m_psv = t.m_psv.ll();
      long m_nsv = t.m_nsv.ll();

      next = parse_phrase(text, length, i, m_psv, m_nsv, pos, len);

      if (write_out) {
        out_writer->write(pos);
        out_writer->write(len);
        out_vol += 2 * sizeof(uint40);
      }
    }

    ++sorter;

    if (dbg == (1 << 23)) {
      dbg = 0;

      long double elapsed = utils::wclock() - parse_start;
      long double scanned_m = i / (1024.L * 1024);
      long inp_vol = sizeof(PsvNsvTriple) * i;
      long tot_vol = inp_vol + out_vol;
      long double tot_vol_m = tot_vol / (1024.L * 1024);
      long double io_speed = tot_vol_m / elapsed;
      fprintf(stderr, "Parsing: position = %.1LfM (%.1Lf%%), "
          "time = %.1Lfs (%.3Lfs/iMiB), io = %.2LfMiB/s\r",
          scanned_m, (100.L * i) / length, elapsed,
          elapsed / scanned_m, io_speed);
    }
  }

  sorter.finish_clear();
  delete out_writer;

  long double parse_time = utils::wclock() - parse_start;
  fprintf(stderr, "\nParsing: time = %.1Lfs (%.3Lfs/iMiB), "
      "max-alloc = %.2Lfbytes/char\n",
      parse_time, parse_time / length_MiB, 1.L * max_alloc() / length);

  long double total_time = utils::wclock() - start;
  fprintf(stderr, "Total time: %.2Lfs (%.4Lfs/iMiB)\n",
      total_time, total_time / length_MiB);

  fprintf(stderr, "Number of phrases = %ld\n", nphrase);
}

#endif // __PARSE_H
